%***************************************************************************
%           JMBC PRACTICUM NUMERIEKE STROMINGSLEER 1998
%
% OPGAVE :   2
% NAAM MODULE: ns_2.m
% DATUM : januari 1998 
%***************************************************************************
% OMSCHRIJVING:
%   Deze module lost een warmtetransportvergelijking op
%            dT/dt + u dT/dx = k d2T/dx2
%   voor 0<x<1 en 0<T<4. Randvoorwaarden zijn T(0,t)=0 en dT/dx(1,t)=0.
%   Beginvoorwaarde T(x,0)=x.
%   Discretisatie methode is gegen. Crank-Nicolson met parameter omega.
%
%**************************************************************************
% INVOER:   NX       aantal roosterpunten in x-richting
%           dt       tijdstap
%           omega    gewichtsfactor in tijdsintegratie
%           u        convectieve snelheid
%           k        diffusiecoefficient 
% UITVOER:  tijd(i)  tijd t=(i-1)*dt, i=1:NTMAX+1
%           temp(i)  T(1,tijd(i))      temperatuur op x=1
%           T(i,n)   T(x(i),tijd(n))   volledige oplossing
%***************************************************************************
%
clear                      % alle variabelen verwijderen 
%
% begin invoer
%
u=input('input velocity u:   ');
k=input('input diffusion k:    ');
NX=input('input grid points NX:   ');
omega=input('input parameter omega:   ');
dt=input('input time step dt:   ');
tmax=4;                       % lengte tijdsinterval
%
% einde invoer 
%
NTMAX=round(tmax/dt);         % aantal tijdstappen
hx=1/NX;                      % maaswijdte
d=2*k*dt/(hx*hx); 
eta = u*dt/hx;                % CFL getal
%
% berekenen matrix coefficienten
%
for i=1:NX-1,
   di(i)=1+omega*d;
   lo(i)=omega*0.5*(-eta-d);
   up(i)=omega*0.5*(eta-d);
end
%
% op x=1 een Neumann randvoorwaarde
%
di(NX)=1+omega*d;
lo(NX)=-omega*d;
up(NX)=0;
%
% vullen matrix 
%
A=diag(up(1:NX-1),1)+diag(di)+diag(lo(2:NX),-1);
% eenmalig A inverteren
AINV=inv(A);
%
% beginvoorwaarde invullen
%
x=[1:NX]*hx;
tnew=x;
%
T=tnew;
tijd(1)=0;                    % begintijdstip
temp(1)=1;                    % beginvoorwaarde
%
% uitvoeren tijdstappen -------------------------------------------
%
for N=1:NTMAX,                % start tijdstappen
   told=tnew;
%
% berekenen rechterlid
%
   rl(1)=(1-omega)*0.5*(-2*d*told(1)+(d-eta)*told(2)) + told(1);
%
   for i=2:NX-1,
      rl(i)=(1-omega)*0.5*((d+eta)*told(i-1)-2*d*told(i)+(d-eta)*told(i+1)) + told(i);
   end
%
% op x=1 een Neumann randvoorwaarde
%
   rl(NX)=(1-omega)*d*(told(NX-1)-told(NX))+told(NX);
%
% oplossen
%
   tnew=AINV*rl';
%
%antwoord opslaan
%
   tijd(N+1)=N*dt;            % nummering is een versprongen
   temp(N+1)=tnew(NX);        % nummering is een versprongen
   T=[T;[tnew(1:NX)']];
end                           % einde tijdstappen   
%
%-----------------------------------------------------------------
%                             
clf                           % begin plaatjes
subplot(121)
plot(tijd,temp)
set(gca,'XLim',[0 tmax]);
title('oplossing op x=1')
xlabel('t')
ylabel('Temp')
tekst=sprintf('dt=%6.4f',dt);
text(0.5,0.8,tekst)
tekst=sprintf('omega=%4.2f',omega);
text(2.7,0.8,tekst)
%
subplot(122)
meshc(tijd,x,T')
set(gca,'XLim',[0 tmax]);
title('Temp(x,t)')
xlabel('t')
ylabel('x')
zlabel('Temp')
view(30,30)
