%*****************************************************************************
%        JMBC PRACTICUM NUMERIEKE STROMINGSLEER 1998
%                  4e-orde methodes
%
% NAAM:      ns_4.m
% DATUM:     januari 1998
%*****************************************************************************
% OMSCHRIJVING:
%   Deze module bepaalt een discrete oplossing van
%         dy/dx - k d2y/dx2 = RL
%   op het interval 0<x<1, met randvoorwaarden y(0)=0 en y(1)=1.
%   Er wordt gediscretiseerd op een gerekt rooster.
%
%   Diverse 2e en 4e-orde methoden zijn geimplementeerd
%*****************************************************************************
% INVOER:   N         initial number of grid points
%                     refinement strategy
%           k         diffusion coefficient
%
%*****************************************************************************
clear              % remove all variables

%
N0=input('input smallest N: ');

hoe=input('increase N with factor or with increment? (0/1):  ');
if hoe==0,
  increm=input('with which factor should N be increased? :   ');
else
  increm=input('with which increment should N be increased? :  ');
end

aantal=input('how often should N be increased? (0/1/2/...):  ');  

k=input('input k: ');

%rechts=input('do you want a right-hand side? (0/1):  ');
rechts=0;

% convectieve snelheid
u=1;

iew=input('do you want eigenvalues? (0/1):  ');
%iew=0;

% berekening van gerekte rooster; N/2 mazen links en rechts van x = 1 - Lk
L=10;

%expro=input('rooster abrupt / exponentieel? (0/1):  ');
%F=input('welk deel van punten links? :  ');
F=0.5;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% begin loop met roosterverfijning

for teller=1:(aantal+1),

for expro = 0:1,          % both abrupt and exponential grid

if hoe==0,
  N = round(N0 * increm^(teller-1));
else
  N = N0 + round((teller-1)*increm);
end

% protection against too fine grids
if N > 320,
   disp(['N = ',num2str(N),' is too large (N_max=320)']);
   return
end
NN(teller)=N;

clear x xrek yrekex b hcrs hfin sqrth Nc Nf fac
clear M2L rl2L y2L M2S rl2S y2S upp low dia
clear M4L rl4L y4L lagm2 lagm1 lag0 lagp1 lagp2
clear M4S rl4S y4S m4Sm2 m4Sm1 m4S0 m4Sp1 m4Sp2 b4S


%%%%%%%%%%%%%%%%%%%%%% grid generation %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

Nc=round(N*F);
Nf=N-Nc;
if Nf <= 3,
  disp(['only ',num2str(Nf),' grid cells in boundary layer!']);
end  
  
% abrupt, piecewise uniform grid

hcrs=(1.-L*k)/Nc;           % grove mazen links
hfin=L*k/Nf;                 % fijne mazen rechts
xrek(1:Nc)=hcrs*(1:1:Nc);                 % roosterpunten links
xrek(Nc+1:N)=xrek(Nc) + hfin*(1:1:Nf);   % roosterpunten rechts
if L*k > 0.5
   hfin=1/N;
   hcrs=1/N;
   xrek(1:N)=(1:1:N)/N;
end
%     
% exponential grid
%
if (expro == 1 & L*k <= 0.5),
% rekking zodanig dat zelfde verdeling als boven 
   fac=(L*k/(1-L*k))^(1/Nc);
%   disp(['rekkingsfactor= ',num2str(fac)])
   mesh=(1-fac)/(1-fac^N);
   xrek(1)=mesh;
   for i=2:N
     mesh=fac*mesh;
     xrek(i)=xrek(i-1)+mesh;
   end
end
%
% mesh sizes ***************************************
%
h(1)=xrek(1);
for i=2:N,
  h(i)=xrek(i)-xrek(i-1);
end

% gewichtsfactor voor norm
for i=1:N-1,
  sqrth(i)=sqrt(0.5*(h(i)+h(i+1)));
end
sqrth(N)=0;
sqrth=sqrth';

%%%%%%%%%%%%%%%%%%%%%%% determine exact solution %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 
% exacte oplossing op gerekt rooster 
%
pe=1/k;
% aktie om underflow van exp(-pe) te voorkomen
if pe > 100,
   empe=0;
else
   empe=exp(-pe);
end
yexrek=1-(1-exp(pe*(xrek-1)))/(1-empe);
%
b = zeros(1,N-1);

if rechts ==1,
%% inhomogeen rechterlid
  for i=1:N-1
    if xrek(i) <= 0.4 
      b(i)=2.-5.*xrek(i);
    else
      b(i)=0;
    end
    yexrek(i)=ns_exact(xrek(i),k);
  end  
  yexrek(N)=1;
end % inhomogeen 
yexrek=yexrek';
 
%%%%%%%%%%%%%%%%%%%%  2e orde methoden  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Tweede orde Lagrange discretisatie
%
for i=1:N-1
   upp(i) = (h(i)* u - 2*k) / h(i+1);
   low(i) = (-h(i+1) * u - 2*k) / h(i);
   dia(i) = -upp(i)-low(i);
   rl2L(i) = b(i) * (h(i)+h(i+1));   
end
%
% matrix
M2L=diag(upp(1:N-2),1) + diag(dia(1:N-1),0) + diag(low(2:N-1),-1);
%
% boundary conditions
rl2L(N-1)=rl2L(N-1)-upp(N-1);
%
% solution
y2L = M2L\rl2L';
y2L(N) = 1;
if iew==1,
   [V2L,D2L]=eig(M2L);
end

subplot(221)
plot(xrek,y2L,xrek,yexrek)
set(gca,'XLim',[0 1])
E2L=norm(sqrth.*(yexrek-y2L));
trek=sprintf('error=%6.1e',E2L);
text(0.1,0.7,trek)
title('2nd Lagrange');
     

%
% 2nd order symmetry-preserving method **********
%
for i=1:N-1
   upp(i) = u - 2*k/h(i+1);
   low(i) = -u - 2*k/h(i);
   dia(i) = -upp(i)-low(i);
   rl2S(i) = b(i) * (h(i)+h(i+1));
end
%
% matrix
M2S=diag(upp(1:N-2),1) + diag(dia(1:N-1),0) + diag(low(2:N-1),-1);
% 
% boundary conditions
rl2S(N-1)=rl2S(N-1)-upp(N-1);
%
% solution
y2S = M2S\rl2S';
y2S(N) = 1;

if iew==1,
%   [V2S,D2S]=eig(M2S);
end

subplot(222)
plot(xrek,y2S,xrek,yexrek)
set(gca,'XLim',[0 1])
E2S=norm(sqrth.*(yexrek-y2S));
trek=sprintf('error=%6.1e',E2S);
text(0.1,0.7,trek)
title('2nd order SP');


%%%%%%%%%%%   4th-order methods %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

x=xrek;
x(N+1)=2*x(N)-x(N-1);

for i=1:N-1

   if i==1,
      xm2=-2*x(1);
      xm1=-x(1);
   elseif i==2,
      xm2=-x(2);
      xm1=x(1)-x(2);
   else
      xm2=x(i-2)-x(i);
      xm1=x(i-1)-x(i);
   end
   xp2=x(i+2)-x(i);
   xp1=x(i+1)-x(i);

%  traditional Lagrangian 4-th order

   noemer = (xm2-xm1)*xm2*(xm2-xp1)*(xm2-xp2);
   cm2 = -xm1*xp1*xp2 / noemer;
   dm2 = 2*(xp1*xp2+xm1*xp1+xm1*xp2) / noemer;
   lagm2(i) = u*cm2 - k*dm2;

   noemer = (xm1-xm2)*xm1*(xm1-xp1)*(xm1-xp2);
   cm1 = -xm2*xp1*xp2 / noemer;
   dm1 = 2*(xp1*xp2+xm2*xp1+xm2*xp2) / noemer;
   lagm1(i) = u*cm1 - k*dm1;

   noemer = (xp1-xm2)*(xp1-xm1)*xp1*(xp1-xp2);
   cp1 = -xm2*xm1*xp2 / noemer;
   dp1 = 2*(xm1*xp2+xm2*xm1+xm2*xp2) / noemer;
   lagp1(i) = u*cp1 - k*dp1;

   noemer = (xp2-xm2)*(xp2-xm1)*xp2*(xp2-xp1);
   cp2 = -xm2*xm1*xp1 / noemer;
   dp2 = 2*(xm1*xp1+xm2*xp1+xm2*xm1) / noemer;
   lagp2(i) = u*cp2 - k*dp2;

   lag0(i) = -lagm2(i) -lagm1(i) -lagp1(i) -lagp2(i);

%  modern symmetry-preserving 4th order 

   noemer4 = -xp2 + 8*xp1 -8*xm1 +xm2;
   b4S(i)=noemer4*b(i);
   m4Sm2(i)=(u-2*k/xm2);
   m4Sm1(i)=-8*(u-2*k/xm1);
   m4Sp1(i)=8*(u-2*k/xp1);
   m4Sp2(i)=-(u-2*k/xp2);
   m4S0(i)=  -m4Sm2(i) -m4Sm1(i) -m4Sp1(i) -m4Sp2(i);

end

% traditional 4th order Lagrangian discretization ***************************

% exact boundary conditions

ym1=1-(1-exp(pe*(-x(1)-1)))/(1-empe);
yNp1=1-(1-exp(pe*(x(N+1)-1)))/(1-empe);

if rechts == 1,
   ym1=ns_exact(-x(1),k);
   yNp1=ns_exact(x(N+1),k);
end

rl4L=b;
rl4L(1)=rl4L(1)-lagm2(1)*ym1;
rl4L(N-2)=rl4L(N-2)-lagp2(N-2);
rl4L(N-1)=rl4L(N-1)-lagp2(N-1)*yNp1-lagp1(N-1);

% matrix
M4L=diag(lagp2(1:N-3),2)+diag(lagp1(1:N-2),1)+diag(lag0(1:N-1),0)+ ...
             diag(lagm1(2:N-1),-1)+diag(lagm2(3:N-1),-2);

% solution	 
y4L=M4L\rl4L';
y4L(N)=1.0;

if iew==1,
   [V4L,D4L]=eig(M4L);
end

subplot(223)
plot(xrek,y4L,xrek,yexrek)
set(gca,'XLim',[0 1])
E4L=norm(sqrth.*(yexrek-y4L));
trek=sprintf('error=%6.1e',E4L);
text(0.1,0.7,trek)
title('4th Lagrange + exact BC');
	 
% symmetry preserving method ************************************

% matrix
M4S=diag(m4Sp2(1:N-3),2)+diag(m4Sp1(1:N-2),1)+diag(m4S0(1:N-1),0)+ ...
         diag(m4Sm1(2:N-1),-1)+diag(m4Sm2(3:N-1),-2);

% laatste punten tweede orde 
     
M4S(1,1)=dia(1);
M4S(1,2)=upp(1);
M4S(1,3)=0;
M4S(N-1,N-3)=0;
M4S(N-1,N-2)=low(N-1);
M4S(N-1,N-1)=dia(N-1);

% boundary conditions 
rl4S=b4S;
rl4S(N-2)=rl4S(N-2)-m4Sp2(N-2);
rl4S(1)=rl2S(1);
rl4S(N-1)=rl2S(N-1);

% solution
y4S=M4S\rl4S';
y4S(N)=1.0;

if iew==1,
%   [V4S,D4S]=eig(M4S);
end

subplot(224)
plot(xrek,y4S,xrek,yexrek)
set(gca,'XLim',[0 1])
E4S=norm(sqrth.*(yexrek-y4S));
trek=sprintf('error=%6.1e',E4S);
text(0.1,0.7,trek)
title('4th order SP + 2nd BC')



%%%%%%%%%%%%%%%%%%%  overview of results %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
 

% number of `negative' eigenvalues

if iew==1,
  t2L=0;
  t2S=0;
  t4L=0;
  t4S=0;

  for i=1:N-1,
    re2L(i)=real(D2L(i,i));
    re4L(i)=real(D4L(i,i))/N;
    if re2L(i) < 0.0,
       t2L=t2L+1;
    end
    if re4L(i) < 0.0,
       t4L=t4L+1;
    end
    im2L(i)=imag(D2L(i,i));
    im4L(i)=imag(D4L(i,i))/N;
  end
end


% store global errors and eigenvalues

if expro == 0,
  soort=(['abrupt grid d=10k:   ']);
else 
  soort=(['exponential d=10k:   ']);  
end

globaal10(teller,1+expro)=norm(sqrth.*(y2L-yexrek));
globaal10(teller,3+expro)=norm(sqrth.*(y2S-yexrek));
globaal10(teller,5+expro)=norm(sqrth.*(y4L-yexrek));
globaal10(teller,7+expro)=norm(sqrth.*(y4S-yexrek));

if iew==1,
  remin10(teller,1+expro)=t2L;
  remin10(teller,3+expro)=t2S;
  remin10(teller,5+expro)=t4L;
  remin10(teller,7+expro)=t4S;
end

% display errors in command window

if expro ==0,
   disp(['  ']);
   disp(['                    ', ...
         '  N','    E2L    ','   E2S    ','   E4L  ','  E4S']);
end

disp([soort, num2str(N),'  ',num2str(E2L),'  ',num2str(E2S),'  ', ...
            num2str(E4L),'  ',num2str(E4S)])
pause

% discretization error as a function of x
%
%subplot(111)
%semilogy(xrek,abs(y2L-yexrek),'r--',xrek,abs(y2S-yexrek),'g--', ...
%         xrek,abs(y4L-yexrek),'r-', xrek,abs(y4S-yexrek),'g-')
%set(gca,'XLim',[0 1],'YLim',[1e-6 1]);
%legend('2L','2S','4L','4S',2);
%pause 
%clf

% display eigenvalues
if iew==1,
  subplot(111)
  plot(re2L*k,im2L*k,'o',re4L*k,im4L*k,'+')
  hold on
  plot([0 0],[-0.2 0.2],'r');
  set(gca,'XLim',[-0.5 1.5], 'YLim',[-0.2 0.2]);
  legend('2L','4L');
  title('eigenvalues of Lagrangian methods');
  zoom on
  hold off
  pause
  zoom off
end    


% end of loop over abrupt or exponential grid
end

% end of loop with grid refinement
end



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

clf

% overview grid refinement

 subplot(221)
 b = loglog(1./NN,abs(globaal10(:,1)),'r--', ... 
        1./NN,abs(globaal10(:,3)),'g--', ...
        1./NN,abs(globaal10(:,5)),'r-', ...
        1./NN,abs(globaal10(:,7)),'g-' );
    
 axis([3e-3,0.1,1e-7,1000]);
 set(gca,'Xtick',[0.01 0.1]);
 legend('2L','2S','4L','4S',2)
 title('Abrupt: d=10k','FontSize',12);    
 xlabel('mean mesh size','Fontsize',12)
 set(b(2),'LineWidth',2);
 set(b(4),'Linewidth',2);


 subplot(222)
 c = loglog(1./NN,abs(globaal10(:,2)),'r--', ...
         1./NN,abs(globaal10(:,4)),'g--', ...
	 1./NN,abs(globaal10(:,6)),'r-', ... 
         1./NN,abs(globaal10(:,8)),'g-'  );
			  
 axis([3e-3,0.1,1e-7,1000]);
 set(gca,'Xtick',[0.01 0.1]);
 title('Exponential: d=10k','Fontsize',12)
 xlabel('mean mesh size','Fontsize',12)
 set(c(2),'LineWidth',2);
 set(c(4),'Linewidth',2);

% eigenvalue information ************************************************

if iew==1,
   subplot(223)				      
   semilogx(1./NN,remin10(:,1),'r--', ...
            1./NN,remin10(:,5),'r-');
   axis([3e-3 0.1 0 15]);
   set(gca,'Xtick',[0.01 0.1]);
   title('Abrupt: d=10k','Fontsize',12)
   xlabel('mean mesh size', 'Fontsize',12);
   ylabel('# eigenvalues < 0','Fontsize',12);
   legend('2L','4L');
   
   subplot(224)
   semilogx(1./NN,remin10(:,2),'r--', ...
         1./NN,remin10(:,6),'r-' );
   axis([3e-3 0.1 0 15]);
   set(gca,'Xtick',[0.01 0.1]);
   title('Exponential: d=10k','Fontsize',12)
   xlabel('mean mesh size', 'Fontsize',12);
end

%   print -deps overview.eps
